@@ -21,20 +21,14 @@
 
 BUILDERS = {
     'source-tarball':        'source_tarball',
-    'msvc2008-win32':        'msvc',
-    'msvc2008-win64':        'msvc',
-    'msvc2010-win32':        'msvc',
-    'msvc2010-win64':        'msvc',
-    'msvc2012-win32':        'msvc',
-    'msvc2012-win64':        'msvc',
     'msvc2013-win32':        'msvc',
     'msvc2013-win64':        'msvc',
-    'msvc-winsdk71-win32':   'msvc_winsdk71',
-    'msvc-winsdk71-win64':   'msvc_winsdk71',
-    'setup-mingw-w64':       'setup_mingw64',
+    'setup-mingw_w64':       'setup_schroot',
     'setup-schroot-centos5': 'setup_schroot',
     'setup-schroot-centos6': 'setup_schroot',
+    'setup-schroot-centos7': 'setup_schroot',
     'setup-schroot-wheezy':  'setup_schroot',
+    'setup-schroot-jessie':  'setup_schroot',
     'setup-schroot-trusty':  'setup_schroot',
     'setup-schroot-precise': 'setup_schroot',
     'update-all-schroots':   'update_schroot',
@@ -42,8 +36,11 @@ BUILDERS = {
     'centos5-amd64':         'linux_schroot',
     'centos6-i386':          'linux_schroot',
     'centos6-amd64':         'linux_schroot',
+    'centos7-amd64':         'linux_schroot',
     'wheezy-i386':           'linux_schroot',
     'wheezy-amd64':          'linux_schroot',
+    'jessie-i386':           'linux_schroot',
+    'jessie-amd64':          'linux_schroot',
     'trusty-i386':           'linux_schroot',
     'trusty-amd64':          'linux_schroot',
     'precise-i386':          'linux_schroot',
@@ -186,8 +183,17 @@ FPM_SETUP = {
         '--provides':        'wkhtmltopdf',
         '--conflicts':       'wkhtmltopdf',
         '--replaces':        'wkhtmltopdf',
-        '--depends':         ['fontconfig', 'libfontconfig1', 'libfreetype6', 'libpng12-0', 'zlib1g', 'libjpeg8',
-                              'libssl1.0.0', 'libx11-6', 'libxext6', 'libxrender1', 'libstdc++6', 'libc6']
+        '--depends':         ['fontconfig', 'libfontconfig1', 'libfreetype6', 'libpng12-0', 'zlib1g', 'libjpeg8', 'libssl1.0.0',
+                              'libx11-6', 'libxext6', 'libxrender1', 'xfonts-base', 'xfonts-75dpi', 'libstdc++6', 'libc6']
+    },
+    'jessie': {
+        '-t':                'deb',
+        '--deb-compression': 'xz',
+        '--provides':        'wkhtmltopdf',
+        '--conflicts':       'wkhtmltopdf',
+        '--replaces':        'wkhtmltopdf',
+        '--depends':         ['fontconfig', 'libfontconfig1', 'libfreetype6', 'libpng12-0', 'zlib1g', 'libjpeg62-turbo', 'libssl1.0.0',
+                              'libx11-6', 'libxext6', 'libxrender1', 'xfonts-base', 'xfonts-75dpi', 'libstdc++6', 'libc6']
     },
     'trusty': {
         '-t':                'deb',
@@ -195,8 +201,8 @@ FPM_SETUP = {
         '--provides':        'wkhtmltopdf',
         '--conflicts':       'wkhtmltopdf',
         '--replaces':        'wkhtmltopdf',
-        '--depends':         ['fontconfig', 'libfontconfig1', 'libfreetype6', 'libpng12-0', 'zlib1g', 'libjpeg-turbo8',
-                              'libssl1.0.0', 'libx11-6', 'libxext6', 'libxrender1', 'libstdc++6', 'libc6']
+        '--depends':         ['fontconfig', 'libfontconfig1', 'libfreetype6', 'libpng12-0', 'zlib1g', 'libjpeg-turbo8', 'libssl1.0.0',
+                              'libx11-6', 'libxext6', 'libxrender1', 'xfonts-base', 'xfonts-75dpi', 'libstdc++6', 'libc6']
     },
     'precise': {
         '-t':                'deb',
@@ -204,22 +210,29 @@ FPM_SETUP = {
         '--provides':        'wkhtmltopdf',
         '--conflicts':       'wkhtmltopdf',
         '--replaces':        'wkhtmltopdf',
-        '--depends':         ['fontconfig', 'libfontconfig1', 'libfreetype6', 'libpng12-0', 'zlib1g', 'libjpeg8',
-                              'libssl1.0.0', 'libx11-6', 'libxext6', 'libxrender1', 'libstdc++6', 'libc6']
+        '--depends':         ['fontconfig', 'libfontconfig1', 'libfreetype6', 'libpng12-0', 'zlib1g', 'libjpeg8', 'libssl1.0.0',
+                              'libx11-6', 'libxext6', 'libxrender1', 'xfonts-base', 'xfonts-75dpi', 'libstdc++6', 'libc6']
     },
     'centos5': {
         '-t':                'rpm',
         '--epoch':           '1',
         '--rpm-compression': 'bzip2',
-        '--depends':         ['fontconfig', 'freetype', 'libpng', 'zlib', 'libjpeg', 'openssl',
-                              'libX11', 'libXext', 'libXrender', 'libstdc++', 'glibc']
+        '--depends':         ['fontconfig', 'freetype', 'libpng', 'zlib', 'libjpeg', 'openssl', 'libstdc++', 'glibc',
+                              'libX11', 'libXext', 'libXrender', 'xorg-x11-fonts-Type1', 'xorg-x11-fonts-75dpi']
     },
     'centos6': {
         '-t':                'rpm',
         '--epoch':           '1',
         '--rpm-compression': 'bzip2',
-        '--depends':         ['fontconfig', 'freetype', 'libpng', 'zlib', 'libjpeg', 'openssl',
-                              'libX11', 'libXext', 'libXrender', 'libstdc++', 'glibc']
+        '--depends':         ['fontconfig', 'freetype', 'libpng', 'zlib', 'libjpeg', 'openssl', 'libstdc++', 'glibc',
+                              'libX11', 'libXext', 'libXrender', 'xorg-x11-fonts-Type1', 'xorg-x11-fonts-75dpi']
+    },
+    'centos7': {
+        '-t':                'rpm',
+        '--epoch':           '1',
+        '--rpm-compression': 'xz',
+        '--depends':         ['fontconfig', 'freetype', 'libpng', 'zlib', 'libjpeg-turbo', 'openssl', 'libstdc++', 'glibc',
+                              'libX11', 'libXext', 'libXrender', 'xorg-x11-fonts-Type1', 'xorg-x11-fonts-75dpi']
     },
     'osx': {
         '-t':                         'osxpkg',
@@ -241,12 +254,28 @@ deb http://security.debian.org/   wheezy/updates main contrib non-free"""),
         ('shell', 'apt-get dist-upgrade --assume-yes'),
         ('shell', 'apt-get install --assume-yes xz-utils libssl-dev libpng-dev libjpeg8-dev zlib1g-dev rubygems'),
         ('shell', 'apt-get install --assume-yes libfontconfig1-dev libfreetype6-dev libx11-dev libxext-dev libxrender-dev'),
-        ('shell', 'gem install fpm ronn --no-ri --no-rdoc'),
-        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\n'),
+        ('shell', 'gem install fpm --no-ri --no-rdoc'),
+        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\ngem update fpm\n'),
         ('fpm_setup',  'fpm_package.sh'),
         ('schroot_conf', 'Debian Wheezy')
     ],
 
+    'jessie': [
+        ('debootstrap', 'jessie', 'http://ftp.us.debian.org/debian/'),
+        ('write_file', 'etc/apt/sources.list', """
+deb http://ftp.debian.org/debian/ jessie         main contrib non-free
+deb http://ftp.debian.org/debian/ jessie-updates main contrib non-free
+deb http://security.debian.org/   jessie/updates main contrib non-free"""),
+        ('shell', 'apt-get update'),
+        ('shell', 'apt-get dist-upgrade --assume-yes'),
+        ('shell', 'apt-get install --assume-yes xz-utils libssl-dev libpng-dev libjpeg62-turbo-dev zlib1g-dev rubygems ruby-dev'),
+        ('shell', 'apt-get install --assume-yes libfontconfig1-dev libfreetype6-dev libx11-dev libxext-dev libxrender-dev'),
+        ('shell', 'gem install fpm --no-ri --no-rdoc'),
+        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\ngem update fpm\n'),
+        ('fpm_setup',  'fpm_package.sh'),
+        ('schroot_conf', 'Debian Jessie')
+    ],
+
     'trusty': [
         ('debootstrap', 'trusty', 'http://archive.ubuntu.com/ubuntu/'),
         ('write_file', 'etc/apt/sources.list', """
@@ -257,14 +286,16 @@ deb http://archive.ubuntu.com/ubuntu/ trusty-security main restricted universe m
         ('shell', 'apt-get dist-upgrade --assume-yes'),
         ('shell', 'apt-get install --assume-yes xz-utils libssl-dev libpng-dev libjpeg-turbo8-dev zlib1g-dev ruby-dev'),
         ('shell', 'apt-get install --assume-yes libfontconfig1-dev libfreetype6-dev libx11-dev libxext-dev libxrender-dev'),
-        ('shell', 'gem install fpm ronn --no-ri --no-rdoc'),
-        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\n'),
+        ('shell', 'gem install fpm --no-ri --no-rdoc'),
+        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\ngem update fpm\n'),
         ('fpm_setup',  'fpm_package.sh'),
         ('schroot_conf', 'Ubuntu Trusty')
     ],
 
     'precise': [
         ('debootstrap', 'precise', 'http://archive.ubuntu.com/ubuntu/'),
+        ('write_file', 'usr/sbin/policy-rc.d', "#!/bin/bash\nexit 101\n"),
+        ('shell', 'chmod a+x /usr/sbin/policy-rc.d'),
         ('write_file', 'etc/apt/sources.list', """
 deb http://archive.ubuntu.com/ubuntu/ precise          main restricted universe multiverse
 deb http://archive.ubuntu.com/ubuntu/ precise-updates  main restricted universe multiverse
@@ -273,8 +304,8 @@ deb http://archive.ubuntu.com/ubuntu/ precise-security main restricted universe
         ('shell', 'apt-get dist-upgrade --assume-yes'),
         ('shell', 'apt-get install --assume-yes xz-utils libssl-dev libpng-dev libjpeg8-dev zlib1g-dev rubygems'),
         ('shell', 'apt-get install --assume-yes libfontconfig1-dev libfreetype6-dev libx11-dev libxext-dev libxrender-dev'),
-        ('shell', 'gem install fpm ronn --no-ri --no-rdoc'),
-        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\n'),
+        ('shell', 'gem install fpm --no-ri --no-rdoc'),
+        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\ngem update fpm\n'),
         ('fpm_setup',  'fpm_package.sh'),
         ('schroot_conf', 'Ubuntu Precise')
     ],
@@ -288,8 +319,8 @@ deb http://archive.ubuntu.com/ubuntu/ precise-security main restricted universe
         ('append_file:amd64', 'etc/yum.conf', 'exclude = *.i?86\n'),
         ('shell', 'yum install -y gcc gcc-c++ make diffutils perl ruby-devel rubygems rpm-build libffi-devel'),
         ('shell', 'yum install -y openssl-devel libX11-devel libXrender-devel libXext-devel fontconfig-devel freetype-devel libjpeg-devel libpng-devel zlib-devel'),
-        ('shell', 'gem install fpm ronn --no-ri --no-rdoc'),
-        ('write_file', 'update.sh', 'yum update -y\n'),
+        ('shell', 'gem install fpm --no-ri --no-rdoc'),
+        ('write_file', 'update.sh', 'yum update -y\ngem update fpm\n'),
         ('fpm_setup',  'fpm_package.sh'),
         ('schroot_conf', 'CentOS 5')
     ],
@@ -300,21 +331,51 @@ deb http://archive.ubuntu.com/ubuntu/ precise-security main restricted universe
         ('append_file:amd64', 'etc/yum.conf', 'exclude = *.i?86\n'),
         ('shell', 'yum install -y gcc gcc-c++ make diffutils perl ruby-devel rubygems rpm-build libffi-devel'),
         ('shell', 'yum install -y openssl-devel libX11-devel libXrender-devel libXext-devel fontconfig-devel freetype-devel libjpeg-devel libpng-devel zlib-devel'),
-        ('shell', 'gem install fpm ronn --no-ri --no-rdoc'),
-        ('write_file', 'update.sh', 'yum update -y\n'),
+        ('shell', 'gem install fpm --no-ri --no-rdoc'),
+        ('write_file', 'update.sh', 'yum update -y\ngem update fpm\n'),
         ('fpm_setup',  'fpm_package.sh'),
         ('schroot_conf', 'CentOS 6')
+    ],
+
+    'centos7:amd64': [
+        ('rinse', 'centos-7'),
+        ('shell', 'yum install -y deltarpm'),
+        ('shell', "sed -i 's/mirrorlist/#mirrorlist/g' /etc/yum.repos.d/CentOS-Base.repo"),
+        ('shell', "sed -i 's/#baseurl/baseurl/g' /etc/yum.repos.d/CentOS-Base.repo"),
+        ('shell', 'yum update -y'),
+        ('append_file', 'etc/yum.conf', 'exclude = *.i?86\n'),
+        ('shell', 'yum install -y gcc gcc-c++ make diffutils perl ruby-devel rubygems rpm-build libffi-devel'),
+        ('shell', 'yum install -y openssl-devel libX11-devel libXrender-devel libXext-devel fontconfig-devel freetype-devel libjpeg-turbo-devel libpng-devel zlib-devel'),
+        ('shell', 'gem install fpm --no-ri --no-rdoc'),
+        ('write_file', 'update.sh', 'yum update -y\ngem update fpm\n'),
+        ('fpm_setup',  'fpm_package.sh'),
+        ('schroot_conf', 'CentOS 7')
+    ],
+
+    'mingw_w64:amd64': [
+        ('set_alias', 'mingw-w64'),
+        ('debootstrap', 'trusty', 'http://archive.ubuntu.com/ubuntu/'),
+        ('write_file', 'etc/apt/sources.list', """
+deb http://archive.ubuntu.com/ubuntu/ trusty          main restricted universe multiverse
+deb http://archive.ubuntu.com/ubuntu/ trusty-updates  main restricted universe multiverse
+deb http://archive.ubuntu.com/ubuntu/ trusty-security main restricted universe multiverse"""),
+        ('shell', 'apt-get update'),
+        ('shell', 'apt-get dist-upgrade --assume-yes'),
+        ('shell', 'apt-get install --assume-yes build-essential mingw-w64 nsis python ruby perl gperf bison flex git'),
+        ('write_file', 'update.sh', 'apt-get update\napt-get dist-upgrade --assume-yes\n'),
+        ('schroot_conf', 'MinGW-w64 on Ubuntu Trusty')
     ]
 }
 
 DEPENDENT_LIBS = {
     'openssl': {
         'order' : 1,
-        'url'   : 'http://www.openssl.org/source/openssl-1.0.1h.tar.gz',
-        'sha1'  : 'b2239599c8bf8f7fc48590a55205c26abe560bf8',
+        'url'   : 'https://www.openssl.org/source/openssl-1.0.1l.tar.gz',
+        'sha1'  : '4547a0b4269acf76b1f9e7d188896867d6fc8c18',
         'build' : {
             'msvc*-win32*': {
                 'result': ['include/openssl/ssl.h', 'lib/ssleay32.lib', 'lib/libeay32.lib'],
+                'replace': [('util/pl/VC-32.pl', ' /MT', ' /MD')],
                 'commands': [
                     'perl Configure --openssldir=%(destdir)s VC-WIN32 no-asm',
                     'ms\\do_ms.bat',
@@ -322,6 +383,7 @@ DEPENDENT_LIBS = {
             },
             'msvc*-win64*': {
                 'result': ['include/openssl/ssl.h', 'lib/ssleay32.lib', 'lib/libeay32.lib'],
+                'replace': [('util/pl/VC-32.pl', ' /MT', ' /MD')],
                 'commands': [
                     'perl Configure --openssldir=%(destdir)s VC-WIN64A',
                     'ms\\do_win64a.bat',
@@ -330,7 +392,7 @@ DEPENDENT_LIBS = {
             'mingw-w64-cross-win*': {
                 'result': ['include/openssl/ssl.h', 'lib/libssl.a', 'lib/libcrypto.a'],
                 'commands': [
-                    'perl Configure --openssldir=%(destdir)s --cross-compile-prefix=%(mingw-w64)s- no-shared no-asm mingw64',
+                    'perl Configure --openssldir=%(destdir)s --cross-compile-prefix=%(mingw_w64)s- no-shared no-asm mingw64',
                     'make',
                     'make install_sw']
             }
@@ -348,7 +410,6 @@ DEPENDENT_LIBS = {
                     'include/zconf.h': 'zconf.h',
                     'lib/zdll.lib'   : 'zlib.lib'
                 },
-                'replace':  [('win32/Makefile.msc', '-MD', '-MT')],
                 'commands': ['nmake /f win32/Makefile.msc zlib.lib']
             },
             'mingw-w64-cross-win*': {
@@ -357,7 +418,7 @@ DEPENDENT_LIBS = {
                     'include/zconf.h': 'zconf.h',
                     'lib/libz.a'     : 'libz.a'
                 },
-                'replace':  [('win32/Makefile.gcc', 'PREFIX =', 'PREFIX = %(mingw-w64)s-')],
+                'replace':  [('win32/Makefile.gcc', 'PREFIX =', 'PREFIX = %(mingw_w64)s-')],
                 'commands': ['make -f win32/Makefile.gcc']
             }
         }
@@ -365,8 +426,8 @@ DEPENDENT_LIBS = {
 
     'libpng': {
         'order' : 3,
-        'url' : 'http://downloads.sourceforge.net/libpng/libpng-1.5.18.tar.gz',
-        'sha1': '7ddf6865aa70d2d79faf65ebc611919a0b573654',
+        'url' : 'http://downloads.sourceforge.net/libpng/libpng-1.5.21.tar.gz',
+        'sha1': '0ce1aa25abd55ad153516a29735b8ad432b4b771',
         'build' : {
             'msvc*': {
                 'result': {
@@ -376,7 +437,6 @@ DEPENDENT_LIBS = {
                     'lib/libpng.lib'      : 'libpng.lib'
                 },
                 'replace': [
-                    ('scripts/makefile.vcwin32', '-MD', '-MT'),
                     ('scripts/makefile.vcwin32', '-I..\\zlib', '-I..\\deplibs\\include'),
                     ('scripts/makefile.vcwin32', '..\\zlib\\zlib.lib', '..\\deplibs\\lib\\zdll.lib')],
                 'commands': ['nmake /f scripts/makefile.vcwin32 libpng.lib']
@@ -391,9 +451,9 @@ DEPENDENT_LIBS = {
                 'replace': [
                     ('scripts/makefile.gcc', 'ZLIBINC = ../zlib', 'ZLIBINC = %(destdir)s/include'),
                     ('scripts/makefile.gcc', 'ZLIBLIB = ../zlib', 'ZLIBLIB = %(destdir)s/lib'),
-                    ('scripts/makefile.gcc', 'CC = gcc', 'CC = %(mingw-w64)s-gcc'),
-                    ('scripts/makefile.gcc', 'AR_RC = ar', 'AR_RC = %(mingw-w64)s-ar'),
-                    ('scripts/makefile.gcc', 'RANLIB = ranlib', 'RANLIB = %(mingw-w64)s-ranlib')],
+                    ('scripts/makefile.gcc', 'CC = gcc', 'CC = %(mingw_w64)s-gcc'),
+                    ('scripts/makefile.gcc', 'AR_RC = ar', 'AR_RC = %(mingw_w64)s-ar'),
+                    ('scripts/makefile.gcc', 'RANLIB = ranlib', 'RANLIB = %(mingw_w64)s-ranlib')],
                 'commands': ['make -f scripts/makefile.gcc libpng.a']
             },
             'osx-carbon-i386': {
@@ -426,7 +486,7 @@ DEPENDENT_LIBS = {
                 },
                 'replace':  [('makefile.vc', '!include <win32.mak>', ''),
                              ('makefile.vc', '$(cc)', 'cl'),
-                             ('makefile.vc', '$(cflags) $(cdebug) $(cvars)', '-c -nologo -D_CRT_SECURE_NO_DEPRECATE -MT -O2 -W3')],
+                             ('makefile.vc', '$(cflags) $(cdebug) $(cvars)', '-c -nologo -D_CRT_SECURE_NO_DEPRECATE -MD -O2 -W3')],
                 'commands': [
                     'copy /y jconfig.vc jconfig.h',
                     'nmake /f makefile.vc libjpeg.lib']
@@ -434,7 +494,7 @@ DEPENDENT_LIBS = {
             'mingw-w64-cross-win*': {
                 'result': ['include/jpeglib.h', 'include/jmorecfg.h', 'include/jerror.h', 'include/jconfig.h', 'lib/libjpeg.a'],
                 'commands': [
-                    './configure --host=%(mingw-w64)s --disable-shared --prefix=%(destdir)s',
+                    './configure --host=%(mingw_w64)s --disable-shared --prefix=%(destdir)s',
                     'make install']
             },
             'osx-carbon-i386': {
@@ -454,13 +514,13 @@ DEPENDENT_LIBS = {
 
     'xz': {
         'order' : 5,
-        'url' : 'http://tukaani.org/xz/xz-5.0.5.tar.gz',
-        'sha1': '26fec2c1e409f736e77a85e4ab314dc74987def0',
+        'url' : 'http://tukaani.org/xz/xz-5.2.0.tar.gz',
+        'sha1': 'ef6b8e0b6fd85eb1b844e555ae3ecbb3a4aa8e81',
         'build' : {
             'osx*': {
                 'result': ['bin/xz'],
                 'commands': [
-                    'CFLAGS="-arch x86_64" ./configure --disable-nls --enable-small --disable-shared --disable-threads --prefix=%(destdir)s',
+                    'CFLAGS="-arch i386 -mmacosx-version-min=10.6" ./configure --disable-nls --enable-small --disable-shared --disable-threads --prefix=%(destdir)s',
                     'make -C src/liblzma', 'make -C src/xz', 'make install-strip']
             }
         }
@@ -503,7 +563,14 @@ import os, sys, platform, subprocess, shutil, re, fnmatch, multiprocessing, urll
 
 from os.path import exists
 
-CPU_COUNT = max(2, multiprocessing.cpu_count()-1)   # leave one CPU free
+if platform.system() == 'Windows':
+    try:
+        import winreg
+    except ImportError:
+        import _winreg as winreg
+    CPU_COUNT = max(2, multiprocessing.cpu_count()-1)   # leave one CPU free
+else:
+    CPU_COUNT = max(2, multiprocessing.cpu_count())
 
 def rchop(s, e):
     if s.endswith(e):
@@ -523,6 +590,11 @@ def shell(cmd):
     if ret != 0:
         error("%s\ncommand failed: exit code %d" % (cmd, ret))
 
+def chroot_shell(name, cmd):
+    ret = os.system('schroot -c wkhtmltopdf-%s -- %s ' % (name, cmd))
+    if ret != 0:
+        error("command inside chroot failed: exit code %d" % ret)
+
 def get_output(*cmd):
     try:
         return subprocess.check_output(cmd, stderr=subprocess.STDOUT).strip()
@@ -539,6 +611,15 @@ def mkdir_p(path):
     if not exists(path):
         os.makedirs(path)
 
+def get_registry_value(key, value=None):
+    for mask in [0, winreg.KEY_WOW64_64KEY, winreg.KEY_WOW64_32KEY]:
+        try:
+            reg_key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, key, 0, winreg.KEY_READ | mask)
+            return winreg.QueryValueEx(reg_key, value)[0]
+        except WindowsError:
+            pass
+    return None
+
 def get_version(basedir):
     mkdir_p(basedir)
     text = open(os.path.join(basedir, '..', 'VERSION'), 'r').read()
@@ -549,7 +630,7 @@ def get_version(basedir):
     hash = get_output('git', 'rev-parse', '--short', 'HEAD')
     if not hash:
         return (text, version)
-    return ('%s-%s' % (version, hash), version)
+    return ('%s-%s' % (text, hash), version)
 
 def qt_config(key, *opts):
     input, output = [], []
@@ -619,11 +700,12 @@ def _is_compiled(dst, loc):
         present = present and exists(os.path.join(dst, name))
     return present
 
-def build_deplibs(config, basedir):
+def build_deplibs(config, basedir, **kwargs):
     mkdir_p(os.path.join(basedir, config))
 
     dstdir = os.path.join(basedir, config, 'deplibs')
-    vars   = {'destdir': dstdir, 'mingw-w64': MINGW_W64_PREFIX.get(rchop(config, '-dbg'), '')}
+    vars   = {'destdir': dstdir }
+    vars.update(kwargs)
     for lib in sorted(DEPENDENT_LIBS, key=lambda x: DEPENDENT_LIBS[x]['order']):
         cfg = None
         for key in DEPENDENT_LIBS[lib]['build']:
@@ -691,13 +773,26 @@ def build_setup_schroot(config, basedir):
 
     login  = os.environ.get('SUDO_USER') or get_output('logname')
     chroot = config[1+config.rindex('-'):]
+
+    command_list = CHROOT_SETUP.get(chroot)
+    if not command_list and ('%s:amd64' % chroot) in CHROOT_SETUP:
+        command_list = CHROOT_SETUP['%s:amd64' % chroot]
+        if 'i386' in ARCH:
+            del ARCH[ARCH.index('i386')]
+
     for arch in ARCH:
-        message('******************* %s-%s\n' % (chroot, arch))
+        alias = '%s-%s' % (chroot, arch)
+        if command_list[0][0] == 'set_alias':
+            alias = command_list[0][1]
+            del command_list[0]
+        message('******************* %s\n' % alias)
         base_dir = os.environ.get('WKHTMLTOX_CHROOT') or '/var/chroot'
-        root_dir = os.path.join(base_dir, 'wkhtmltopdf-%s-%s' % (chroot, arch))
+        root_dir = os.path.join(base_dir, alias)
+        os.system('umount %s/proc' % root_dir)
+        os.system('umount %s/sys'  % root_dir)
         rmdir(root_dir)
         mkdir_p(root_dir)
-        for command in CHROOT_SETUP[chroot]:
+        for command in command_list:
             # handle architecture-specific commands
             name = command[0]
             if ':' in name:
@@ -710,9 +805,14 @@ def build_setup_schroot(config, basedir):
             if name == 'debootstrap':
                 shell('debootstrap --arch=%(arch)s --variant=buildd %(distro)s %(dir)s %(url)s' % {
                     'arch': arch, 'dir': root_dir, 'distro': command[1], 'url': command[2] })
+                cmd = (arch == 'i386' and 'linux32 chroot' or 'chroot')
+                shell('%s %s mount -t proc  proc  /proc' % (cmd, root_dir))
+                shell('%s %s mount -t sysfs sysfs /sys'  % (cmd, root_dir))
             elif name == 'rinse':
                 cmd = (arch == 'i386' and 'linux32 rinse' or 'rinse')
                 shell('%s --arch %s --distribution %s --directory %s' % (cmd, arch, command[1], root_dir))
+                shell('%s %s mount -t proc  proc  /proc' % (cmd, root_dir))
+                shell('%s %s mount -t sysfs sysfs /sys'  % (cmd, root_dir))
             elif name == 'shell':
                 cmd = (arch == 'i386' and 'linux32 chroot' or 'chroot')
                 shell('%s %s %s' % (cmd, root_dir, command[1]))
@@ -736,14 +836,16 @@ def build_setup_schroot(config, basedir):
                 open(loc, 'w').write(cmd % (args, cfg['--name'], chroot, cfg['-t']))
                 shell('chmod a+x %s' % loc)
             elif name == 'schroot_conf':
-                cfg = open('/etc/schroot/chroot.d/wkhtmltopdf-%s-%s' % (chroot, arch), 'w')
-                cfg.write('[wkhtmltopdf-%s-%s]\n' % (chroot, arch))
+                cfg = open('/etc/schroot/chroot.d/wkhtmltopdf-%s' % alias, 'w')
+                cfg.write('[wkhtmltopdf-%s]\n' % alias)
                 cfg.write('type=directory\ndirectory=%s/\n' % root_dir)
                 cfg.write('description=%s %s for wkhtmltopdf\n' % (command[1], arch))
                 cfg.write('users=%s\nroot-users=root\n' % login)
                 if arch == 'i386' and 'amd64' in ARCH:
                     cfg.write('personality=linux32\n')
                 cfg.close()
+        os.system('umount %s/proc' % root_dir)
+        os.system('umount %s/sys'  % root_dir)
 
 def check_update_schroot(config):
     check_running_on_debian()
@@ -755,12 +857,6 @@ def build_update_schroot(config, basedir):
         message('******************* %s\n' % name[name.index('wkhtmltopdf-'):])
         shell('schroot -c %s -- /bin/bash /update.sh' % name[name.index('wkhtmltopdf-'):])
 
-def check_setup_mingw64(config):
-    check_running_on_debian()
-
-def build_setup_mingw64(config, basedir):
-    install_packages('build-essential', 'mingw-w64', 'nsis')
-
 def check_source_tarball(config):
     if not get_output('git', 'rev-parse', '--short', 'HEAD'):
         error("This can only be run inside a git checkout.")
@@ -792,14 +888,15 @@ def build_source_tarball(config, basedir):
         tar.add('.', 'wkhtmltox-%s/' % version, filter=_filter_tar)
     shell('git reset --hard HEAD')
 
-# --------------------------------------------------------------- MSVC (2008-2013)
+# --------------------------------------------------------------- MSVC (2013 only)
 
 MSVC_LOCATION = {
-    'msvc2008': 'VS90COMNTOOLS',
-    'msvc2010': 'VS100COMNTOOLS',
-    'msvc2012': 'VS110COMNTOOLS',
     'msvc2013': 'VS120COMNTOOLS'
 }
+MSVC_RUNTIME = {
+    'msvc2013-win32': ('df7f0a73bfa077e483e51bfb97f5e2eceedfb6a3', 'http://download.microsoft.com/download/2/E/6/2E61CFA4-993B-4DD4-91DA-3737CD5CD6E3/vcredist_x86.exe'),
+    'msvc2013-win64': ('8bf41ba9eef02d30635a10433817dbb6886da5a2', 'http://download.microsoft.com/download/2/E/6/2E61CFA4-993B-4DD4-91DA-3737CD5CD6E3/vcredist_x64.exe')
+}
 
 def check_msvc(config):
     version, arch = rchop(config, '-dbg').split('-')
@@ -818,6 +915,14 @@ def check_msvc(config):
                        and not exists(os.path.join(vcdir, 'bin', 'x86_amd64', 'cl.exe')):
         error("%s: unable to find the amd64 compiler" % version)
 
+    perl = get_output('perl', '-V')
+    if not perl or 'perl5' not in perl:
+        error("perl does not seem to be installed.")
+
+    nsis = get_registry_value(r'SOFTWARE\NSIS')
+    if not nsis or not exists(os.path.join(nsis, 'makensis.exe')):
+        error("NSIS does not seem to be installed.")
+
 def build_msvc(config, basedir):
     msvc, arch = rchop(config, '-dbg').split('-')
     vcdir = os.path.join(os.environ[MSVC_LOCATION[msvc]], '..', '..', 'VC')
@@ -838,46 +943,12 @@ def build_msvc(config, basedir):
 
     os.environ.update(eval(stdout.strip()))
 
-    build_msvc_common(config, basedir)
-
-# --------------------------------------------------------------- MSVC via Windows SDK 7.1
-
-def check_msvc_winsdk71(config):
-    for pfile in ['ProgramFiles(x86)', 'ProgramFiles']:
-        if pfile in os.environ and exists(os.path.join(os.environ[pfile], 'Microsoft SDKs', 'Windows', 'v7.1', 'Bin', 'SetEnv.cmd')):
-            return
-    error("Unable to detect the location of Windows SDK 7.1")
-
-def build_msvc_winsdk71(config, basedir):
-    arch = config[config.rindex('-'):]
-    setenv = None
-    for pfile in ['ProgramFiles(x86)', 'ProgramFiles']:
-        if not pfile in os.environ:
-            continue
-        setenv = os.path.join(os.environ[pfile], 'Microsoft SDKs', 'Windows', 'v7.1', 'Bin', 'SetEnv.cmd')
-
-    mode = debug and '/Debug' or '/Release'
-    if arch == 'win64':
-        args = '/2008 /x64 %s' % mode
-    else:
-        args = '/2008 /x86 %s' % mode
-
-    python = sys.executable
-    process = subprocess.Popen('("%s" %s>nul)&&"%s" -c "import os, sys; sys.stdout.write(repr(dict(os.environ)))"' % (
-        setenv, args, python), stdout=subprocess.PIPE, shell=True)
-    stdout, _ = process.communicate()
-    exitcode = process.wait()
-    if exitcode != 0:
-        error("unable to initialize the environment for Windows SDK 7.1")
-
-    os.environ.update(eval(stdout.strip()))
-
-    build_msvc_common(config, basedir)
-
-def build_msvc_common(config, basedir):
     version, simple_version = get_version(basedir)
     build_deplibs(config, basedir)
 
+    sha1, url = MSVC_RUNTIME[rchop(config, '-dbg')]
+    shutil.copy(download_file(url, sha1, basedir), os.path.join(basedir, config, 'vcredist.exe'))
+
     libdir = os.path.join(basedir, config, 'deplibs')
     qtdir  = os.path.join(basedir, config, 'qt')
     mkdir_p(qtdir)
@@ -904,19 +975,10 @@ def build_msvc_common(config, basedir):
     shell('%s\\bin\\qmake %s\\..\\wkhtmltopdf.pro' % (qtdir, basedir))
     shell('nmake')
 
-    found = False
-    for pfile in ['ProgramFiles(x86)', 'ProgramFiles']:
-        if not pfile in os.environ or not exists(os.path.join(os.environ[pfile], 'NSIS', 'makensis.exe')):
-            continue
-        found = True
-
-        makensis = os.path.join(os.environ[pfile], 'NSIS', 'makensis.exe')
-        os.chdir(os.path.join(basedir, '..'))
-        shell('"%s" /DVERSION=%s /DSIMPLE_VERSION=%s /DTARGET=%s wkhtmltox.nsi' % \
-                (makensis, version, simple_version, config))
-
-    if not found:
-        message("\n\nCould not build installer as NSIS was not found.\n")
+    makensis = os.path.join(get_registry_value(r'SOFTWARE\NSIS'), 'makensis.exe')
+    os.chdir(os.path.join(basedir, '..'))
+    shell('"%s" /DVERSION=%s /DSIMPLE_VERSION=%s /DTARGET=%s /DMSVC /DARCH=%s wkhtmltox.nsi' % \
+            (makensis, version, simple_version, config, arch))
 
 # ------------------------------------------------ MinGW-W64 Cross Environment
 
@@ -926,22 +988,26 @@ MINGW_W64_PREFIX = {
 }
 
 def check_mingw64_cross(config):
-    shell('%s-gcc --version' % MINGW_W64_PREFIX[rchop(config, '-dbg')])
+    chroot_shell('mingw-w64', '%s-gcc --version' % MINGW_W64_PREFIX[rchop(config, '-dbg')])
 
 def build_mingw64_cross(config, basedir):
+    os.chdir(os.path.realpath(os.path.join(basedir, '..')))
+    chroot_shell('mingw-w64', 'python scripts/build.py %s -chroot-build' % ' '.join(sys.argv[1:]))
+
+def chroot_build_mingw64_cross(config, basedir):
     version, simple_version = get_version(basedir)
-    build_deplibs(config, basedir)
+    build_deplibs(config, basedir, mingw_w64=MINGW_W64_PREFIX.get(rchop(config, '-dbg')))
 
     libdir = os.path.join(basedir, config, 'deplibs')
     qtdir  = os.path.join(basedir, config, 'qt')
 
     configure_args = qt_config('mingw-w64-cross',
         '--prefix=%s'   % qtdir,
-        '-I %s/include' % libdir,
-        '-L %s/lib'     % libdir,
+        '-I%s/include'  % libdir,
+        '-L%s/lib'      % libdir,
         '-device-option CROSS_COMPILE=%s-' % MINGW_W64_PREFIX[rchop(config, '-dbg')])
 
-    os.environ['OPENSSL_LIBS'] = '-lssl -lcrypto -L %s/lib -lws2_32 -lgdi32 -lcrypt32' % libdir
+    os.environ['OPENSSL_LIBS'] = '-lssl -lcrypto -L%s/lib -lws2_32 -lgdi32 -lcrypt32' % libdir
 
     mkdir_p(qtdir)
     os.chdir(qtdir)
@@ -964,10 +1030,18 @@ def build_mingw64_cross(config, basedir):
     shell('%s/bin/qmake -spec win32-g++-4.6 %s/../wkhtmltopdf.pro' % (qtdir, basedir))
     shell('make')
     shutil.copy('bin/libwkhtmltox0.a', 'bin/wkhtmltox.lib')
+    shell('rm -f bin/lib*.dll')
+    for dll in ['libgcc_s_sjlj-1.dll', 'libstdc++-6.dll', 'libwinpthread-1.dll']:
+        dll_path = get_output('dpkg', '-S', dll)
+        if dll_path:
+            for line in dll_path.split('\n'):
+                loc = line[1+line.index(':'):].strip()
+                if exists(loc) and MINGW_W64_PREFIX[rchop(config, '-dbg')] in loc:
+                    shell('cp %s bin/' % loc)
 
     os.chdir(os.path.join(basedir, '..'))
-    shell('makensis -DVERSION=%s -DSIMPLE_VERSION=%s -DTARGET=%s wkhtmltox.nsi' % \
-            (version, simple_version, config))
+    shell('makensis -DVERSION=%s -DSIMPLE_VERSION=%s -DTARGET=%s -DMINGW -DARCH=%s wkhtmltox.nsi' % \
+            (version, simple_version, config, rchop(config, '-dbg').split('-')[-1]))
 
 # -------------------------------------------------- Linux schroot environment
 
@@ -1055,7 +1129,7 @@ def build_posix_local(config, basedir):
     shell('cp ../../../include/wkhtmltox/*.h ../wkhtmltox-%s/include/wkhtmltox' % version)
     shell('cp ../../../include/wkhtmltox/dll*.inc ../wkhtmltox-%s/include/wkhtmltox' % version)
 
-    os.chdir(basedir)
+    os.chdir(os.path.join(basedir, config))
     shell('tar -c -v -f ../wkhtmltox-%s_local-%s.tar wkhtmltox-%s/' % (version, platform.node(), version))
     shell('xz --compress --force --verbose -9 ../wkhtmltox-%s_local-%s.tar' % (version, platform.node()))
 
@@ -1192,7 +1266,7 @@ def main():
         usage()
 
     for arg in sys.argv[2:]:
-        if not arg in ['-clean', '-debug']:
+        if not arg in ['-clean', '-debug', '-chroot-build']:
             usage()
 
     final_config = config
@@ -1204,6 +1278,10 @@ def main():
         rmdir(os.path.join(basedir, config))
 
     os.chdir(rootdir)
+    if '-chroot-build' in sys.argv[2:]:
+        globals()['chroot_build_%s' % BUILDERS[config]](final_config, basedir)
+        return
+
     globals()['check_%s' % BUILDERS[config]](final_config)
     globals()['build_%s' % BUILDERS[config]](final_config, basedir)
 
